from code42cli.cmds.detectionlists import (
    DetectionList,
    DetectionListHandlers,
    load_user_descriptions,
    get_user_id,
    update_user,
)


def load_subcommands():
    handlers = _create_handlers()
    detection_list = DetectionList.create_departing_employee_list(handlers)
    return detection_list.load_subcommands()


def _create_handlers():
    return DetectionListHandlers(
        add=add_departing_employee, remove=remove_departing_employee, load_add=_load_add_description
    )


def add_departing_employee(
    sdk, profile, username, cloud_alias=None, departure_date=None, notes=None
):
    """Adds an employee to the departing employee detection list.

    Args:
        sdk (py42.sdk.SDKClient): py42.
        profile (C42Profile): Your code42 profile.
        username (str): The username of the employee to add.
        cloud_alias (str): An alternative email address for another cloud service.
        departure_date (str): The date the employee is departing in format `YYYY-MM-DD`.
        notes: (str): Notes about the employee.
    """
    user_id = get_user_id(sdk, username)
    sdk.detectionlists.departing_employee.add(user_id, departure_date)
    update_user(sdk, user_id, cloud_alias, notes=notes)


def remove_departing_employee(sdk, profile, username):
    user_id = get_user_id(sdk, username)
    sdk.detectionlists.departing_employee.remove(user_id)


def _load_add_description(argument_collection):
    load_user_descriptions(argument_collection)
    departure_date = argument_collection.arg_configs[u"departure_date"]
    departure_date.set_help(u"The date the employee is departing in format YYYY-MM-DD.")



class DetectionLists(object):
    DEPARTING_EMPLOYEE = u"departing-employee"
    HIGH_RISK_EMPLOYEE = u"high-risk-employee"


class BulkCommandType(object):
    ADD = u"add"
    REMOVE = u"remove"

    def __iter__(self):
        return iter([self.ADD, self.REMOVE])


class DetectionListUserKeys(object):
    CLOUD_ALIAS = u"cloud_alias"
    USERNAME = u"username"
    NOTES = u"notes"
    RISK_TAG = u"risk_tag"



IS_INCREMENTAL_KEY = u"incremental"


class OutputFormat(object):
    CEF = "CEF"
    JSON = "JSON"
    RAW = "RAW-JSON"

    def __iter__(self):
        return iter([self.CEF, self.JSON, self.RAW])


class ExposureType(object):
    SHARED_VIA_LINK = "SharedViaLink"
    SHARED_TO_DOMAIN = "SharedToDomain"
    APPLICATION_READ = "ApplicationRead"
    CLOUD_STORAGE = "CloudStorage"
    REMOVABLE_MEDIA = "RemovableMedia"
    IS_PUBLIC = "IsPublic"

    def __iter__(self):
        return iter(self._as_list())

    def __len__(self):
        return len(self._as_list())

    def _as_list(self):
        return [
            self.SHARED_VIA_LINK,
            self.SHARED_TO_DOMAIN,
            self.APPLICATION_READ,
            self.CLOUD_STORAGE,
            self.REMOVABLE_MEDIA,
            self.IS_PUBLIC,
        ]


class ServerProtocol(object):
    TCP = "TCP"
    UDP = "UDP"

    def __iter__(self):
        return iter([self.TCP, self.UDP])


class SearchArguments(object):
    """These string values should match `argparse` stored parameter names. For example, for the 
    CLI argument `--c42-username`, the string should be `c42_username`."""

    ADVANCED_QUERY = u"advanced_query"
    BEGIN_DATE = u"begin"
    END_DATE = u"end"
    EXPOSURE_TYPES = u"type"
    C42_USERNAME = u"c42_username"
    ACTOR = u"actor"
    MD5 = u"md5"
    SHA256 = u"sha256"
    SOURCE = u"source"
    FILE_NAME = u"file_name"
    FILE_PATH = u"file_path"
    PROCESS_OWNER = u"process_owner"
    TAB_URL = u"tab_url"
    INCLUDE_NON_EXPOSURE_EVENTS = u"include_non_exposure"

    def __iter__(self):
        return iter(
            [
                self.ADVANCED_QUERY,
                self.BEGIN_DATE,
                self.END_DATE,
                self.EXPOSURE_TYPES,
                self.C42_USERNAME,
                self.ACTOR,
                self.MD5,
                self.SHA256,
                self.SOURCE,
                self.FILE_NAME,
                self.FILE_PATH,
                self.PROCESS_OWNER,
                self.TAB_URL,
                self.INCLUDE_NON_EXPOSURE_EVENTS,
            ]
        )



from code42cli.cmds.detectionlists.commands import DetectionListCommandFactory
from code42cli.bulk import generate_template, run_bulk_process, CSVReader, FlatFileReader
from code42cli.util import print_error
from code42cli.cmds.detectionlists.enums import (
    BulkCommandType,
    DetectionLists,
    DetectionListUserKeys,
)


class UserDoesNotExistError(Exception):
    """An error to represent a username that is not in our system. The CLI shows this error when 
    the user tries to add or remove a user that does not exist. This error is not shown during 
    bulk add or remove."""

    def __init__(self, username):
        super(UserDoesNotExistError, self).__init__(u"User '{}' does not exist.".format(username))


class DetectionListHandlers(object):
    """Handlers DTO for passing in specific detection list functions.
    
    Args:
        add (callable): A function that adds an employee to the list.
        remove (callable): A function that removes an employee from the list.
        load_add (callable): A function that loads the add-related `ArgConfig`s.
    """

    def __init__(self, add=None, remove=None, load_add=None):
        self.add_employee = add
        self.remove_employee = remove
        self.load_add_description = load_add


class DetectionList(object):
    """An object representing a Code42 detection list. Use this class by passing in handlers for 
    adding and removing employees. This class will handle the bulk-related commands and some 
    shared help texts.
    
    Args:
        list_name (str or unicode): An option from the DetectionLists enum. For convenience, use one of the 
            given `classmethods`.
        handlers (DetectionListHandlers): A DTO containing implementations for adding / removing 
            users from specific lists.
        cmd_factory (DetectionListCommandFactory): A factory that creates detection list commands.
    """

    def __init__(self, list_name, handlers, cmd_factory=None):
        self.name = list_name
        self.handlers = handlers
        self.factory = cmd_factory or DetectionListCommandFactory(list_name)

    @classmethod
    def create_high_risk_employee_list(cls, handlers):
        """Creates a high risk employee detection list.
        
        Args:
            handlers (DetectionListHandlers): A DTO containing implementations for adding / 
                removing users from specific lists.
        
        Returns:
            DetectionList: A high risk employee detection list.
        """
        return cls(DetectionLists.HIGH_RISK_EMPLOYEE, handlers)

    @classmethod
    def create_departing_employee_list(cls, handlers):
        """Creates a departing employee detection list.
        
        Args:
            handlers (DetectionListHandlers): A DTO containing implementations for adding /
                removing users from specific lists.
        
        Returns:
            DetectionList: A departing employee detection list.
        """
        return cls(DetectionLists.DEPARTING_EMPLOYEE, handlers)

    def load_subcommands(self):
        """Loads high risk employee related subcommands"""
        bulk = self.factory.create_bulk_command(lambda: self._load_bulk_subcommands())
        add = self.factory.create_add_command(
            self.handlers.add_employee, self.handlers.load_add_description
        )
        remove = self.factory.create_remove_command(
            self.handlers.remove_employee, load_username_description
        )
        return [bulk, add, remove]

    def _load_bulk_subcommands(self):
        generate_template_cmd = self.factory.create_bulk_generate_template_command(
            self.generate_template_file
        )
        add = self.factory.create_bulk_add_command(self.bulk_add_employees)
        remove = self.factory.create_bulk_remove_command(self.bulk_remove_employees)
        return [generate_template_cmd, add, remove]

    def generate_template_file(self, cmd, path=None):
        """Generates a template file a user would need to fill-in for bulk operating on the 
        detection list.
        
        Args:
            cmd (str or unicode): An option from the `BulkCommandType` enum specifying which type of file to 
                generate.
            path (str or unicode, optional): A path to put the file after it's generated. If None, will use 
                the current working directory. Defaults to None.
        """
        handler = None
        if cmd == BulkCommandType.ADD:
            handler = self.handlers.add_employee
        elif cmd == BulkCommandType.REMOVE:
            handler = self.handlers.remove_employee

        generate_template(handler, path)

    def bulk_add_employees(self, sdk, profile, csv_file):
        """Takes a csv file with each row representing an employee and adds them all to a 
        detection list in a bulk fashion.

        Args:
            sdk (py42.sdk.SDKClient): The py42 sdk.
            profile (Code42Profile): The profile under which to execute this command.
            csv_file (str or unicode): The path to the csv file containing rows of users.
        """
        run_bulk_process(
            csv_file, lambda **kwargs: self._add_employee(sdk, profile, **kwargs), CSVReader()
        )

    def bulk_remove_employees(self, sdk, profile, users_file):
        """Takes a flat file with each row containing a username and removes them all from the 
        detection list in a bulk fashion.
        
        Args:
            sdk (py42.sdk.SDKClient): The py42 sdk.
            profile (Code42Profile): The profile under which to execute this command.
            users_file (str or unicode): The path to the file containing rows of user names.
        """
        run_bulk_process(
            users_file,
            lambda *args, **kwargs: self._remove_employee(sdk, profile, *args, **kwargs),
            FlatFileReader(),
        )

    def _add_employee(self, sdk, profile, **kwargs):
        self.handlers.add_employee(sdk, profile, **kwargs)

    def _remove_employee(self, sdk, profile, *args, **kwargs):
        self.handlers.remove_employee(sdk, profile, *args, **kwargs)



def load_username_description(argument_collection):
    """Loads the arg descriptions for the `username` CLI parameter."""
    username = argument_collection.arg_configs[DetectionListUserKeys.USERNAME]
    username.set_help(u"A code42 username for an employee.")


def load_user_descriptions(argument_collection):
    """Loads the arg descriptions related to updating fields about a detection list user, such as 
    notes or a cloud alias.
    
    Args:
        argument_collection (ArgConfigCollection): The arg configs off the command that needs its 
            user descriptions loaded.
    """
    load_username_description(argument_collection)
    cloud_alias = argument_collection.arg_configs[DetectionListUserKeys.CLOUD_ALIAS]
    notes = argument_collection.arg_configs[DetectionListUserKeys.NOTES]
    cloud_alias.set_help(u"An alternative email address for another cloud service.")
    notes.set_help(u"Notes about the employee.")


def get_user_id(sdk, username):
    """Returns the user's UID (referred to by `user_id` in detection lists). If the user does not 
    exist, it prints an error and exits.
    
    Args:
        sdk (py42.sdk.SDKClient): The py42 sdk.
        username (str or unicode): The username of the user to get an ID for.
    
    Returns:
         str: The user ID for the user with the given username.
    """
    users = sdk.users.get_by_username(username)[u"users"]
    if not users:
        ex = UserDoesNotExistError(username)
        print_error(str(ex))
        raise ex
    return users[0][u"userUid"]


def update_user(sdk, user_id, cloud_alias=None, risk_tag=None, notes=None):
    """Updates a detection list user.
    
    Args:
        user_id (str or unicode): The ID of the user to update. This is their `userUid` found from 
            `sdk.users.get_by_username()`.
        cloud_alias (str or unicode): A cloud alias to add to the user.
        risk_tag (iter[str or unicode]): A list of risk tags associated with user.
        notes (str or unicode): Notes about the user.
    """
    if cloud_alias:
        sdk.detectionlists.add_user_cloud_alias(user_id, cloud_alias)
    if risk_tag:
        sdk.detectionlists.add_user_risk_tags(user_id, risk_tag)
    if notes:
        sdk.detectionlists.update_user_notes(user_id, notes)



from code42cli.cmds.detectionlists.enums import BulkCommandType
from code42cli.commands import Command


def create_usage_prefix(detection_list_name):
    return u"code42 {}".format(detection_list_name)


def create_bulk_usage_prefix(detection_list_name):
    return u"{} bulk".format(create_usage_prefix(detection_list_name))


class DetectionListCommandFactory:
    _USAGE_SUFFIX = u"<username> <optional args>"

    def __init__(self, detection_list_name):
        self._name = detection_list_name
        self._usage_prefix = create_usage_prefix(detection_list_name)
        self._bulk_usage_prefix = create_bulk_usage_prefix(detection_list_name)

    def create_bulk_command(self, subcommand_loader):
        return Command(
            u"bulk",
            u"Tools for executing bulk {} commands.".format(self._name),
            subcommand_loader=subcommand_loader,
        )

    def create_add_command(self, handler, arg_customizer):
        return Command(
            BulkCommandType.ADD,
            u"Add a user to the {} detection list.".format(self._name),
            u"{} {} {}".format(self._usage_prefix, BulkCommandType.ADD, self._USAGE_SUFFIX),
            handler=handler,
            arg_customizer=arg_customizer,
        )

    def create_remove_command(self, handler, arg_customizer):
        return Command(
            BulkCommandType.REMOVE,
            u"Remove a user from the {} detection list.".format(self._name),
            u"{} {} {}".format(self._usage_prefix, BulkCommandType.REMOVE, self._USAGE_SUFFIX),
            handler=handler,
            arg_customizer=arg_customizer,
        )

    def create_bulk_generate_template_command(self, handler):
        return Command(
            u"generate-template",
            u"Generate the necessary csv template needed for bulk adding users.",
            u"{} generate-template <cmd> <optional args>".format(self._bulk_usage_prefix),
            handler=handler,
            arg_customizer=DetectionListCommandFactory._load_bulk_generate_template_description,
        )

    def create_bulk_add_command(self, handler):
        return Command(
            BulkCommandType.ADD,
            u"Bulk add users to the {} detection list using a csv file.".format(self._name),
            u"{} {} <csv-file>".format(self._bulk_usage_prefix, BulkCommandType.ADD),
            handler=handler,
            arg_customizer=self._load_bulk_add_description,
        )

    def create_bulk_remove_command(self, handler):
        return Command(
            BulkCommandType.REMOVE,
            u"Bulk remove users from the {} detection list using a file.".format(self._name),
            u"{} {} <file>".format(self._bulk_usage_prefix, BulkCommandType.REMOVE),
            handler=handler,
            arg_customizer=self._load_bulk_remove_description,
        )

    @staticmethod
    def _load_bulk_generate_template_description(argument_collection):
        cmd_type = argument_collection.arg_configs[u"cmd"]
        cmd_type.set_help(u"The type of command the template with be used for.")
        cmd_type.set_choices(BulkCommandType())

    def _load_bulk_add_description(self, argument_collection):
        csv_file = argument_collection.arg_configs[u"csv_file"]
        csv_file.set_help(
            u"The path to the csv file for bulk adding users to the {} detection list.".format(
                self._name
            )
        )

    def _load_bulk_remove_description(self, argument_collection):
        users_file = argument_collection.arg_configs[u"users_file"]
        users_file.set_help(
            u"A file containing a line-separated list of users to remove form the {} detection list".format(
                self._name
            )
        )



from code42cli.args import ArgConfig
from code42cli.cmds.securitydata import enums, logger_factory
from code42cli.cmds.securitydata.extraction import extract
from code42cli.cmds.shared.cursor_store import FileEventCursorStore
from code42cli.commands import Command


def load_subcommands():
    """Sets up the `security-data` subcommand with all of its subcommands."""
    usage_prefix = u"code42 security-data"

    print_func = Command(
        u"print",
        u"Print file events to stdout",
        u"{} {}".format(usage_prefix, u"print <optional-args>"),
        handler=print_out,
        arg_customizer=_load_search_args,
        use_single_arg_obj=True,
    )

    write = Command(
        u"write-to",
        u"Write file events to the file with the given name.",
        u"{} {}".format(usage_prefix, u"write-to <filename> <optional-args>"),
        handler=write_to,
        arg_customizer=_load_write_to_args,
        use_single_arg_obj=True,
    )

    send = Command(
        u"send-to",
        u"Send file events to the given server address.",
        u"{} {}".format(usage_prefix, u"send-to <server-address> <optional-args>"),
        handler=send_to,
        arg_customizer=_load_send_to_args,
        use_single_arg_obj=True,
    )

    clear = Command(
        u"clear-checkpoint",
        u"Remove the saved checkpoint from 'incremental' (-i) mode.",
        u"{} {}".format(usage_prefix, u"clear-checkpoint <optional-args>"),
        handler=clear_checkpoint,
    )

    return [print_func, write, send, clear]


def clear_checkpoint(sdk, profile):
    """Removes the stored checkpoint that keeps track of the last event you got.
        To use, run `code42 security-data clear-checkpoint`.
        This affects `incremental` mode by causing it to behave like it has never been run before.
    """
    FileEventCursorStore(profile.name).replace_stored_insertion_timestamp(None)


def print_out(sdk, profile, args):
    """Activates 'print' command. It gets security events and prints them to stdout."""
    logger = logger_factory.get_logger_for_stdout(args.format)
    extract(sdk, profile, logger, args)


def write_to(sdk, profile, args):
    """Activates 'write-to' command. It gets security events and writes them to the given file."""
    logger = logger_factory.get_logger_for_file(args.output_file, args.format)
    extract(sdk, profile, logger, args)


def send_to(sdk, profile, args):
    """Activates 'send-to' command. It gets security events and logs them to the given server."""
    logger = logger_factory.get_logger_for_server(args.server, args.protocol, args.format)
    extract(sdk, profile, logger, args)


def _load_write_to_args(arg_collection):
    output_file = ArgConfig(u"output_file", help=u"The name of the local file to send output to.")
    arg_collection.append(u"output_file", output_file)
    _load_search_args(arg_collection)


def _load_send_to_args(arg_collection):
    send_to_args = {
        u"server": ArgConfig(u"server", help=u"The server address to send output to."),
        u"protocol": ArgConfig(
            u"-p",
            u"--protocol",
            choices=enums.ServerProtocol(),
            default=enums.ServerProtocol.UDP,
            help=u"Protocol used to send logs to server.",
        ),
    }

    arg_collection.extend(send_to_args)
    _load_search_args(arg_collection)


def _load_search_args(arg_collection):
    search_args = {
        enums.SearchArguments.ADVANCED_QUERY: ArgConfig(
            u"--{}".format(enums.SearchArguments.ADVANCED_QUERY.replace(u"_", u"-")),
            help=u"A raw JSON file event query. "
            u"Useful for when the provided query parameters do not satisfy your requirements."
            u"WARNING: Using advanced queries ignores all other query parameters.",
        ),
        enums.SearchArguments.BEGIN_DATE: ArgConfig(
            u"-b",
            u"--{}".format(enums.SearchArguments.BEGIN_DATE),
            help=u"The beginning of the date range in which to look for events, "
            u"can be a date/time in YYYY-MM-DD (UTC) or YYYY-MM-DD HH:MM:SS (UTC+24-hr time) format "
            u"or a short value representing days (30d), hours (24h) or minutes (15m) from current "
            u"time.",
        ),
        enums.SearchArguments.END_DATE: ArgConfig(
            u"-e",
            u"--{}".format(enums.SearchArguments.END_DATE),
            help=u"The end of the date range in which to look for events, "
            u"can be a date/time in YYYY-MM-DD (UTC) or YYYY-MM-DD HH:MM:SS (UTC+24-hr time) format "
            u"or a short value representing days (30d), hours (24h) or minutes (15m) from current "
            u"time.",
        ),
        enums.SearchArguments.EXPOSURE_TYPES: ArgConfig(
            u"-t",
            u"--{}".format(enums.SearchArguments.EXPOSURE_TYPES),
            nargs=u"+",
            help=u"Limits events to those with given exposure types. "
            u"Available choices={0}".format(list(enums.ExposureType())),
        ),
        enums.SearchArguments.C42_USERNAME: ArgConfig(
            u"--{}".format(enums.SearchArguments.C42_USERNAME.replace(u"_", u"-")),
            nargs=u"+",
            help=u"Limits events to endpoint events for these users.",
        ),
        enums.SearchArguments.ACTOR: ArgConfig(
            u"--{}".format(enums.SearchArguments.ACTOR),
            nargs=u"+",
            help=u"Limits events to only those enacted by the cloud service user of the person who caused the event.",
        ),
        enums.SearchArguments.MD5: ArgConfig(
            u"--{}".format(enums.SearchArguments.MD5),
            nargs=u"+",
            help=u"Limits events to file events where the file has one of these MD5 hashes.",
        ),
        enums.SearchArguments.SHA256: ArgConfig(
            u"--{}".format(enums.SearchArguments.SHA256),
            nargs=u"+",
            action=u"store",
            help=u"Limits events to file events where the file has one of these SHA256 hashes.",
        ),
        enums.SearchArguments.SOURCE: ArgConfig(
            u"--{}".format(enums.SearchArguments.SOURCE),
            nargs=u"+",
            help=u"Limits events to only those from one of these sources. Example=Gmail.",
        ),
        enums.SearchArguments.FILE_NAME: ArgConfig(
            u"--{}".format(enums.SearchArguments.FILE_NAME.replace(u"_", u"-")),
            nargs=u"+",
            help=u"Limits events to file events where the file has one of these names.",
        ),
        enums.SearchArguments.FILE_PATH: ArgConfig(
            u"--{}".format(enums.SearchArguments.FILE_PATH.replace(u"_", u"-")),
            nargs=u"+",
            help=u"Limits events to file events where the file is located at one of these paths.",
        ),
        enums.SearchArguments.PROCESS_OWNER: ArgConfig(
            u"--{}".format(enums.SearchArguments.PROCESS_OWNER.replace(u"_", u"-")),
            nargs=u"+",
            help=u"Limits events to exposure events where one of these users "
            u"owns the process behind the exposure.",
        ),
        enums.SearchArguments.TAB_URL: ArgConfig(
            u"--{}".format(enums.SearchArguments.TAB_URL.replace(u"_", u"-")),
            nargs=u"+",
            help=u"Limits events to be exposure events with one of these destination tab URLs.",
        ),
        enums.SearchArguments.INCLUDE_NON_EXPOSURE_EVENTS: ArgConfig(
            u"--include-non-exposure",
            action=u"store_true",
            help=u"Get all events including non-exposure events.",
        ),
        u"format": ArgConfig(
            u"-f",
            u"--format",
            choices=enums.OutputFormat(),
            default=enums.OutputFormat.JSON,
            help=u"The format used for outputting events.",
        ),
        u"incremental": ArgConfig(
            u"-i",
            u"--incremental",
            action=u"store_true",
            help=u"Only get events that were not previously retrieved.",
        ),
    }

    arg_collection.extend(search_args)



from code42cli.cmds.detectionlists import (
    DetectionList,
    DetectionListHandlers,
    load_user_descriptions,
    load_username_description,
    get_user_id,
    update_user,
)
from code42cli.cmds.detectionlists.enums import DetectionListUserKeys
from code42cli.commands import Command


def load_subcommands():

    handlers = _create_handlers()
    detection_list = DetectionList.create_high_risk_employee_list(handlers)
    cmd_list = detection_list.load_subcommands()
    cmd_list.extend(
        [
            Command(
                u"add-risk-tags",
                u"Associates risk tags with a user.",
                handler=add_risk_tags,
                arg_customizer=_load_risk_tag_mgmt_descriptions,
            ),
            Command(
                u"remove-risk-tags",
                u"Disassociates risk tags from a user.",
                handler=remove_risk_tags,
                arg_customizer=_load_risk_tag_mgmt_descriptions,
            ),
        ]
    )
    return cmd_list


def _create_handlers():
    return DetectionListHandlers(
        add=add_high_risk_employee, remove=remove_high_risk_employee, load_add=_load_add_description
    )


def add_risk_tags(sdk, profile, username, risk_tag):
    risk_tag = _handle_list_args(risk_tag)
    user_id = get_user_id(sdk, username)
    sdk.detectionlists.add_user_risk_tags(user_id, risk_tag)


def remove_risk_tags(sdk, profile, username, risk_tag):
    risk_tag = _handle_list_args(risk_tag)
    user_id = get_user_id(sdk, username)
    sdk.detectionlists.remove_user_risk_tags(user_id, risk_tag)


def add_high_risk_employee(sdk, profile, username, cloud_alias=None, risk_tag=None, notes=None):
    """Adds an employee to the high risk employee detection list.
    
    Args:
        sdk (py42.sdk.SDKClient): py42.
        profile (C42Profile): Your code42 profile.
        username (str): The username of the employee to add.
        cloud_alias (str): An alternative email address for another cloud service.
        risk_tag (iter[str]): Risk tags associated with the employee.
        notes: (str): Notes about the employee.
    """
    risk_tag = _handle_list_args(risk_tag)
    user_id = get_user_id(sdk, username)
    sdk.detectionlists.high_risk_employee.add(user_id)
    update_user(sdk, user_id, cloud_alias, risk_tag, notes)


def remove_high_risk_employee(sdk, profile, username):
    """Removes an employee from the high risk employee detection list.
    
    Args:
        sdk (py42.sdk.SDKClient): py42.
        profile (C42Profile): Your code42 profile.
        username (str): The username of the employee to remove.
    """
    user_id = get_user_id(sdk, username)
    sdk.detectionlists.high_risk_employee.remove(user_id)


def _load_risk_tag_description(argument_collection):
    risk_tag = argument_collection.arg_configs[DetectionListUserKeys.RISK_TAG]
    risk_tag.as_multi_val_param()
    risk_tag.set_help(
        u"Risk tags associated with the employee. "
        u"Options include "
        u"[HIGH_IMPACT_EMPLOYEE, "
        u"ELEVATED_ACCESS_PRIVILEGES, "
        u"PERFORMANCE_CONCERNS, "
        u"FLIGHT_RISK, "
        u"SUSPICIOUS_SYSTEM_ACTIVITY, "
        u"POOR_SECURITY_PRACTICES, "
        u"CONTRACT_EMPLOYEE]"
    )


def _load_add_description(argument_collection):
    load_user_descriptions(argument_collection)
    _load_risk_tag_description(argument_collection)


def _load_risk_tag_mgmt_descriptions(argument_collection):
    load_username_description(argument_collection)
    _load_risk_tag_description(argument_collection)


def _handle_list_args(list_arg):
    """Converts str args to a list. Useful for `bulk` commands which don't use `argparse` but 
    instead pass in values from files, such as in the form "item1 item2"."""
    if list_arg and type(list_arg) != list:
        return list_arg.split()
    return list_arg



import platform
import sys

from py42.settings import set_user_agent_suffix

from code42cli import PRODUCT_NAME
from code42cli.cmds import profile
from code42cli.cmds.detectionlists import departing_employee as de
from code42cli.cmds.detectionlists import high_risk_employee as hre
from code42cli.cmds.detectionlists.enums import DetectionLists
from code42cli.cmds.securitydata import main as secmain
from code42cli.commands import Command
from code42cli.invoker import CommandInvoker

# If on Windows, configure console session to handle ANSI escape sequences correctly
# source: https://bugs.python.org/issue29059
if platform.system().lower() == u"windows":
    from ctypes import windll, c_int, byref

    stdout_handle = windll.kernel32.GetStdHandle(c_int(-11))
    mode = c_int(0)
    windll.kernel32.GetConsoleMode(c_int(stdout_handle), byref(mode))
    mode = c_int(mode.value | 4)
    windll.kernel32.SetConsoleMode(c_int(stdout_handle), mode)


# Sets part of the user agent string that py42 attaches to requests for the purposes of
# identifying CLI users.
set_user_agent_suffix(PRODUCT_NAME)


def main():
    top = Command(u"", u"", subcommand_loader=_load_top_commands)
    invoker = CommandInvoker(top)
    invoker.run(sys.argv[1:])


def _load_top_commands():
    detection_lists_description = u"For adding and removing employees from the {} detection list."

    return [
        Command(
            u"profile", u"For managing Code42 settings.", subcommand_loader=profile.load_subcommands
        ),
        Command(
            u"security-data",
            u"Tools for getting security related data, such as file events.",
            subcommand_loader=secmain.load_subcommands,
        ),
        Command(
            DetectionLists.DEPARTING_EMPLOYEE,
            detection_lists_description.format(u"departing employee"),
            subcommand_loader=de.load_subcommands,
        ),
        Command(
            DetectionLists.HIGH_RISK_EMPLOYEE,
            detection_lists_description.format(u"high risk employee"),
            subcommand_loader=hre.load_subcommands,
        ),
    ]


if __name__ == u"__main__":
    main()



import code42cli.password as password
from code42cli.cmds.shared.cursor_store import get_file_event_cursor_store
from code42cli.config import ConfigAccessor, config_accessor, NoConfigProfileError
from code42cli.util import (
    print_error,
    print_create_profile_help,
    print_set_default_profile_help,
    print_no_existing_profile_message,
)


class Code42Profile(object):
    def __init__(self, profile):
        self._profile = profile

    @property
    def name(self):
        return self._profile.name

    @property
    def authority_url(self):
        return self._profile[ConfigAccessor.AUTHORITY_KEY]

    @property
    def username(self):
        return self._profile[ConfigAccessor.USERNAME_KEY]

    @property
    def ignore_ssl_errors(self):
        return self._profile[ConfigAccessor.IGNORE_SSL_ERRORS_KEY]

    @property
    def has_stored_password(self):
        stored_password = password.get_stored_password(self)
        return stored_password is not None and stored_password != u""

    def get_password(self):
        pwd = password.get_stored_password(self)
        if not pwd:
            pwd = password.get_password_from_prompt()
        return pwd

    def __str__(self):
        return u"{0}: Username={1}, Authority URL={2}".format(
            self.name, self.username, self.authority_url
        )


def _get_profile(profile_name=None):
    """Returns the profile for the given name."""
    config_profile = config_accessor.get_profile(profile_name)
    return Code42Profile(config_profile)


def get_profile(profile_name=None):
    if profile_name is None:
        validate_default_profile()
    try:
        return _get_profile(profile_name)
    except NoConfigProfileError as ex:
        print_error(str(ex))
        print_create_profile_help()
        exit(1)


def default_profile_exists():
    try:
        profile = _get_profile()
        return profile.name and profile.name != ConfigAccessor.DEFAULT_VALUE
    except NoConfigProfileError:
        return False


def is_default_profile(name):
    if default_profile_exists():
        default = get_profile()
        return name == default.name


def validate_default_profile():
    if not default_profile_exists():
        existing_profiles = get_all_profiles()
        if not existing_profiles:
            print_no_existing_profile_message()
        else:
            print_set_default_profile_help(existing_profiles)
        exit(1)


def profile_exists(profile_name=None):
    try:
        _get_profile(profile_name)
        return True
    except NoConfigProfileError:
        return False


def switch_default_profile(profile_name):
    config_accessor.switch_default_profile(profile_name)


def create_profile(name, server, username, ignore_ssl_errors):
    if profile_exists(name):
        print_error(u"A profile named {} already exists.".format(name))
        exit(1)

    config_accessor.create_profile(name, server, username, ignore_ssl_errors)


def delete_profile(profile_name):
    profile = _get_profile(profile_name)
    if password.get_stored_password(profile) is not None:
        password.delete_password(profile)
    cursor_store = get_file_event_cursor_store(profile_name)
    cursor_store.clean()
    config_accessor.delete_profile(profile_name)


def update_profile(name, server, username, ignore_ssl_errors):
    config_accessor.update_profile(name, server, username, ignore_ssl_errors)


def get_all_profiles():
    profiles = [Code42Profile(profile) for profile in config_accessor.get_all_profiles()]
    return profiles


def get_stored_password(profile_name=None):
    profile = get_profile(profile_name)
    return password.get_stored_password(profile)


def set_password(new_password, profile_name=None):
    profile = get_profile(profile_name)
    password.set_password(profile, new_password)



from __future__ import print_function, with_statement

import sys
from os import makedirs, path

from code42cli.compat import open


def get_input(prompt):
    """Uses correct input function based on Python version."""
    # pylint: disable=undefined-variable
    if sys.version_info >= (3, 0):
        return input(prompt)
    else:
        return raw_input(prompt)


def does_user_agree(prompt):
    """Prompts the user and checks if they said yes."""
    ans = get_input(prompt)
    ans = ans.strip().lower()
    return ans == u"y"


def get_user_project_path(subdir=u""):
    """The path on your user dir to /.code42cli/[subdir]."""
    package_name = __name__.split(u".")[0]
    home = path.expanduser(u"~")
    hidden_package_name = u".{0}".format(package_name)
    user_project_path = path.join(home, hidden_package_name, subdir)
    if not path.exists(user_project_path):
        makedirs(user_project_path)
    return user_project_path


def open_file(file_path, mode, action):
    """Wrapper for opening files, useful for testing purposes."""
    with open(file_path, mode, encoding=u"utf-8") as f:
        return action(f)


def print_error(error_text):
    """Prints red text."""
    print(u"\033[91mERROR: {}\033[0m".format(error_text))


def print_bold(bold_text):
    print(u"\033[1m{}\033[0m".format(bold_text))


def is_interactive():
    return sys.stdin.isatty()


def print_no_existing_profile_message():
    print_error(u"No existing profile.")
    print_create_profile_help()


def print_create_profile_help():
    print(u"\nTo add a profile, use: ")
    print_bold(u"\tcode42 profile create <profile-name> <authority-URL> <username>\n")


def print_set_default_profile_help(existing_profiles):
    print(
        u"\nNo default profile set.\n",
        u"\nUse the --profile flag to specify which profile to use.\n",
        u"\nTo set the default profile (used whenever --profile argument is not provided), use:",
    )
    print_bold(u"\tcode42 profile use <profile-name>")
    print(u"\nExisting profiles:")
    for profile in existing_profiles:
        print("\t{}".format(profile))
    print(u"")


def get_url_parts(url_str):
    parts = url_str.split(u":")
    port = None
    if len(parts) > 1 and parts[1] != u"":
        port = int(parts[1])
    return parts[0], port


def print_to_stderr(error_text):
    sys.stderr.write(error_text)



from __future__ import print_function

import sys

from py42.exceptions import Py42ForbiddenError

from code42cli.parser import ArgumentParserError, CommandParser
from code42cli.errors import log_error
from code42cli.util import print_error


class CommandInvoker(object):
    def __init__(self, top_command, cmd_parser=None):
        self._top_command = top_command
        self._cmd_parser = cmd_parser or CommandParser()
        self._commands = {u"": self._top_command}

    def run(self, input_args):
        """Locates a command that matches the one specified by
        `input_args` and runs it with the supplied parameters.
        
        Args:
            input_args (iter[str]): the full list of arguments
            supplied by the user to `code42` cli command.
        """
        try:
            path_parts = self._get_path_parts(input_args)
            command = self._commands.get(u" ".join(path_parts))
            self._try_run_command(command, path_parts, input_args)
        except Py42ForbiddenError as err:
            log_error(err)
            print_error(
                u"You do not have the necessary permissions to perform this task. "
                u"Try using or creating a different profile."
            )
        except Exception as ex:
            log_error(ex)

    def _get_path_parts(self, input_args):
        """Gets the portion of `input_args` that refers to a
        valid command or subcommand, removing parameters.
        For example, `input_args` of ["command", "sub", "--arg", "argval"]
        would return ["command", "sub"], assuming "command" is a top level command,
        "sub" is a subcommand of "command", and the rest of the values are normal parameters.
        Returns an empty string if a valid command or subcommand is not found.
        """
        path = u""
        node = self._commands[u""]
        # step through each segment of input_args until we find
        # something that _isn't_ a command or subcommand.
        for arg in input_args:
            new_path = u"{} {}".format(path, arg).strip()
            self._load_subcommands(path, node)
            node = self._commands.get(new_path)
            if not node:
                break
            path = new_path
        return path.split()

    def _load_subcommands(self, path, node):
        """Discovers a command's subcommands and registers them
        to the available list of commands for this Invoker."""
        node.load_subcommands()
        for command in node.subcommands:
            new_key = u"{} {}".format(path, command.name).strip()
            self._commands[new_key] = command

    def _try_run_command(self, command, path_parts, input_args):
        """Runs a command called using `path_parts` by parsing
        `input_args` and calling the command's handler."""
        try:
            if not path_parts:
                parser = self._cmd_parser.prepare_cli_help(command)
            else:
                parser = self._cmd_parser.prepare_command(command, path_parts)
            parsed_args = self._cmd_parser.parse_args(input_args)
            parsed_args.func(parsed_args)
        except ArgumentParserError as e:
            print(u"error: {}".format(e), file=sys.stderr)
            parser.print_help(sys.stderr)
            sys.exit(2)



from py42.sdk.queries.fileevents.filters.event_filter import EventTimestamp

from code42cli.date_helper import DateArgumentException, parse_min_timestamp, parse_max_timestamp


def create_event_timestamp_filter(begin_date=None, end_date=None):
    """Creates a `py42.sdk.file_event_query.event_query.EventTimestamp` filter using the given dates.
        Returns None if not given a begin_date or an end_date.
        Args:
            begin_date: The begin date for the range.
            end_date: The end date for the range.
    """
    if begin_date and end_date:
        min_timestamp = parse_min_timestamp(begin_date)
        max_timestamp = parse_max_timestamp(end_date)
        return _create_in_range_filter(min_timestamp, max_timestamp)

    elif begin_date and not end_date:
        min_timestamp = parse_min_timestamp(begin_date)
        return _create_on_or_after_filter(min_timestamp)

    elif end_date and not begin_date:
        max_timestamp = parse_max_timestamp(end_date)
        return _create_on_or_before_filter(max_timestamp)


def _create_in_range_filter(min_timestamp, max_timestamp):
    _verify_timestamp_order(min_timestamp, max_timestamp)
    return EventTimestamp.in_range(min_timestamp, max_timestamp)


def _verify_timestamp_order(min_timestamp, max_timestamp):
    if min_timestamp is None or max_timestamp is None:
        return
    if min_timestamp >= max_timestamp:
        raise DateArgumentException(u"Begin date cannot be after end date")


def _create_on_or_after_filter(min_timestamp):
    return EventTimestamp.on_or_after(min_timestamp)


def _create_on_or_before_filter(max_timestamp):
    return EventTimestamp.on_or_before(max_timestamp)



from code42cli.logger import get_error_logger, ERROR_LOG_FILE_NAME
from code42cli.util import is_interactive, print_error, get_user_project_path


ERRORED = False


def log_error(exception):
    """Logs the error to the CLI error log file. If running interactively, it will also print a 
    message telling the user the location of the error log file."""
    logger = get_error_logger()
    logger.error(exception)
    global ERRORED
    ERRORED = True
    print_errors_occurred_if_needed()


def print_errors_occurred_if_needed():
    """If interactive and errors occurred, it will print a message telling the user how to retrieve 
    error logs."""
    if is_interactive() and ERRORED:
        print_errors_occurred()


def print_errors_occurred():
    """Prints a message telling the user how to retrieve error logs."""
    print_error(get_error_message())


def get_error_message():
    """Returns the error message that is printed when errors occur."""
    path = get_user_project_path(u"log")
    return u"View exceptions that occurred at {}/{}.".format(path, ERROR_LOG_FILE_NAME)



import logging
from logging.handlers import RotatingFileHandler
from threading import Lock

from code42cli.compat import str
from code42cli.util import get_user_project_path, print_error


logger_deps_lock = Lock()
ERROR_LOG_FILE_NAME = u"code42_errors.log"


def get_error_logger():
    """Gets the logger where exceptions are logged."""
    log_path = get_user_project_path(u"log")
    log_path = u"{}/{}".format(log_path, ERROR_LOG_FILE_NAME)
    logger = logging.getLogger(u"code42_error_logger")
    if logger_has_handlers(logger):
        return logger

    with logger_deps_lock:
        if not logger_has_handlers(logger):
            formatter = logging.Formatter(u"%(asctime)s %(message)s")
            handler = RotatingFileHandler(log_path, maxBytes=250000000, encoding=u"utf-8")
            return apply_logger_dependencies(logger, handler, formatter)
    return logger


def logger_has_handlers(logger):
    return len(logger.handlers)


def apply_logger_dependencies(logger, handler, formatter):
    try:
        handler.setFormatter(formatter)
        logger.addHandler(handler)
    except Exception as ex:
        print_error(str(ex))
        exit(1)
    return logger



from __future__ import print_function

import json

from c42eventextractor import FileEventHandlers
from c42eventextractor.extractors import FileEventExtractor
from py42.sdk.queries.fileevents.filters import *

import code42cli.cmds.securitydata.date_helper as date_helper
from code42cli.cmds.securitydata.enums import (
    ExposureType as ExposureTypeOptions,
    IS_INCREMENTAL_KEY,
    SearchArguments,
)
from code42cli.logger import get_error_logger
from code42cli.cmds.shared.cursor_store import FileEventCursorStore
from code42cli.compat import str
from code42cli.util import is_interactive, print_bold, print_error, print_to_stderr
import code42cli.errors as errors


_TOTAL_EVENTS = 0


def extract(sdk, profile, output_logger, args):
    """Extracts file events using the given command-line arguments.

        Args:
            sdk (py42.sdk.SDKClient): The py42 sdk.
            profile (Code42Profile): The profile under which to execute this command.
            output_logger (Logger): The logger specified by which subcommand you use. For example,
                print: uses a logger that streams to stdout.
                write-to: uses a logger that logs to a file.
                send-to: uses a logger that sends logs to a server.
            args: Command line args used to build up file event query filters.
    """
    store = _create_cursor_store(args, profile)
    filters = _get_filters(args, store)
    handlers = _create_event_handlers(output_logger, store)
    extractor = FileEventExtractor(sdk, handlers)
    _call_extract(extractor, filters, args.advanced_query)
    _handle_result()


def _create_cursor_store(args, profile):
    if args.incremental:
        return FileEventCursorStore(profile.name)


def _get_filters(args, cursor_store):
    if not _determine_if_advanced_query(args):
        _verify_begin_date_requirements(args, cursor_store)
        _verify_exposure_types(args.type)
        return _create_filters(args)
    else:
        return args.advanced_query


def _determine_if_advanced_query(args):
    if args.advanced_query is not None:
        given_args = vars(args)
        for key in given_args:
            val = given_args[key]
            if not _verify_compatibility_with_advanced_query(key, val):
                print_error(u"You cannot use --advanced-query with additional search args.")
                exit(1)
        return True
    return False


def _verify_begin_date_requirements(args, cursor_store):
    if _begin_date_is_required(args, cursor_store) and not args.begin:
        print_error(u"'begin date' is required.")
        print(u"")
        print_bold(u"Try using  '-b' or '--begin'. Use `-h` for more info.")
        print(u"")
        exit(1)


def _begin_date_is_required(args, cursor_store):
    if not args.incremental:
        return True
    is_required = cursor_store and cursor_store.get_stored_insertion_timestamp() is None

    # Ignore begin date when is incremental mode, it is not required, and it was passed an argument.
    if not is_required and args.begin:
        args.begin = None
    return is_required


def _verify_exposure_types(exposure_types):
    if exposure_types is None:
        return
    options = list(ExposureTypeOptions())
    for exposure_type in exposure_types:
        if exposure_type not in options:
            print_error(u"'{0}' is not a valid exposure type.".format(exposure_type))
            exit(1)


def _create_filters(args):
    filters = []
    event_timestamp_filter = _get_event_timestamp_filter(args.begin, args.end)
    not event_timestamp_filter or filters.append(event_timestamp_filter)
    not args.c42_username or filters.append(DeviceUsername.is_in(args.c42_username))
    not args.actor or filters.append(Actor.is_in(args.actor))
    not args.md5 or filters.append(MD5.is_in(args.md5))
    not args.sha256 or filters.append(SHA256.is_in(args.sha256))
    not args.source or filters.append(Source.is_in(args.source))
    not args.file_name or filters.append(FileName.is_in(args.file_name))
    not args.file_path or filters.append(FilePath.is_in(args.file_path))
    not args.process_owner or filters.append(ProcessOwner.is_in(args.process_owner))
    not args.tab_url or filters.append(TabURL.is_in(args.tab_url))
    _try_append_exposure_types_filter(filters, args.include_non_exposure, args.type)
    return filters


def _get_event_timestamp_filter(begin_date, end_date):
    try:
        begin_date = begin_date.strip() if begin_date else None
        end_date = end_date.strip() if end_date else None
        return date_helper.create_event_timestamp_filter(begin_date, end_date)
    except date_helper.DateArgumentException as ex:
        print_error(str(ex))
        exit(1)


def _create_event_handlers(output_logger, cursor_store):
    handlers = FileEventHandlers()
    error_logger = get_error_logger()

    def handle_error(exception):
        error_logger.error(exception)
        errors.ERRORED = True

    handlers.handle_error = handle_error

    if cursor_store:
        handlers.record_cursor_position = cursor_store.replace_stored_insertion_timestamp
        handlers.get_cursor_position = cursor_store.get_stored_insertion_timestamp

    def handle_response(response):
        response_dict = json.loads(response.text)
        events = response_dict.get(u"fileEvents")
        global _TOTAL_EVENTS
        _TOTAL_EVENTS += len(events)
        for event in events:
            output_logger.info(event)

    handlers.handle_response = handle_response
    return handlers


def _call_extract(extractor, filters, advanced_query):
    if advanced_query:
        extractor.extract_advanced(advanced_query)
    else:
        extractor.extract(*filters)


def _verify_compatibility_with_advanced_query(key, val):
    if key == SearchArguments.INCLUDE_NON_EXPOSURE_EVENTS and not val:
        return True

    if val is not None:
        is_other_search_arg = key in SearchArguments() and key != SearchArguments.ADVANCED_QUERY
        is_incremental = key == IS_INCREMENTAL_KEY and val
        return not is_other_search_arg and not is_incremental
    return True


def _handle_result():
    # Have to call this explicitly (instead of relying on invoker) because errors are caught in
    # `c42eventextractor`.
    errors.print_errors_occurred_if_needed()
    if not _TOTAL_EVENTS:
        print_to_stderr(u"No results found\n")


def _try_append_exposure_types_filter(filters, include_non_exposure_events, exposure_types):
    _exposure_filter = _create_exposure_type_filter(include_non_exposure_events, exposure_types)
    if _exposure_filter:
        filters.append(_exposure_filter)


def _create_exposure_type_filter(include_non_exposure_events, exposure_types):
    if include_non_exposure_events and exposure_types:
        print_error(u"Cannot use exposure types with `--include-non-exposure`.")
        exit(1)
    if exposure_types:
        return ExposureType.is_in(exposure_types)
    if not include_non_exposure_events:
        return ExposureType.exists()
